/*
 *
 * Distributed under the MIT License, available in the file LICENSE.
 *
 * Author: Pavel Kirienko <pavel.kirienko@zubax.com>
 */

#include "canard_stm32.h"
#include "_internal_bxcan.h"
#include "stm32f4xx.h"
#include "Que.h"
//#include <unistd.h>
#define USERXINT 1
#if USERXINT
#define MAXRXFRAMES	10
static uint8_t RxBuffer[PortCANLast][1+(MAXRXFRAMES*sizeof(CanardCANFrame ))];
Q_t CANRxQ[PortCANLast]={
		{.Buffer=RxBuffer[0],.Len=sizeof(RxBuffer[0])},
		{.Buffer=RxBuffer[1],.Len=sizeof(RxBuffer[0])},
		{.Buffer=RxBuffer[2],.Len=sizeof(RxBuffer[0])},
};
#endif
static void usleep(uint16_t i)
{
  while(i--)
  {
  asm("NOP");
  asm("NOP");
  asm("NOP");
  asm("NOP");
  }
};

static volatile CanardSTM32CANType* const bxCAN[PortCANLast] = {CANARD_STM32_CAN1,CANARD_STM32_CAN2,CANARD_STM32_CAN3};
/*
 * State variables
 */
CanardSTM32Stats g_stats[PortCANLast];

static bool g_abort_tx_on_error = false;


static bool isFramePriorityHigher(uint32_t a, uint32_t b)
{
    const uint32_t clean_a = a & CANARD_CAN_EXT_ID_MASK;
    const uint32_t clean_b = b & CANARD_CAN_EXT_ID_MASK;

    /*
     * STD vs EXT - if 11 most significant bits are the same, EXT loses.
     */
    const bool ext_a = (a & CANARD_CAN_FRAME_EFF) != 0;
    const bool ext_b = (b & CANARD_CAN_FRAME_EFF) != 0;
    if (ext_a != ext_b)
    {
        const uint32_t arb11_a = ext_a ? (clean_a >> 18) : clean_a;
        const uint32_t arb11_b = ext_b ? (clean_b >> 18) : clean_b;
        if (arb11_a != arb11_b)
        {
            return arb11_a < arb11_b;
        }
        else
        {
            return ext_b;
        }
    }

    /*
     * RTR vs Data frame - if frame identifiers and frame types are the same, RTR loses.
     */
    const bool rtr_a = (a & CANARD_CAN_FRAME_RTR) != 0;
    const bool rtr_b = (b & CANARD_CAN_FRAME_RTR) != 0;
    if ((clean_a == clean_b) && (rtr_a != rtr_b))
    {
        return rtr_b;
    }

    /*
     * Plain ID arbitration - greater value loses.
     */
    return clean_a < clean_b;
}

/// Converts libcanard ID value into the bxCAN TX ID register format.
static uint32_t convertFrameIDCanardToRegister(const uint32_t id)
{
    uint32_t out = 0;

    if (id & CANARD_CAN_FRAME_EFF)
    {
        out = ((id & CANARD_CAN_EXT_ID_MASK) << 3) | CANARD_STM32_CAN_TIR_IDE;
    }
    else
    {
        out = ((id & CANARD_CAN_STD_ID_MASK) << 21);
    }

    if (id & CANARD_CAN_FRAME_RTR)
    {
        out |= CANARD_STM32_CAN_TIR_RTR;
    }

    return out;
}

/// Converts bxCAN TX/RX (sic! both RX/TX are supported) ID register value into the libcanard ID format.
static uint32_t convertFrameIDRegisterToCanard(const uint32_t id)
{
#if (CANARD_STM32_CAN_TIR_RTR != CANARD_STM32_CAN_RIR_RTR) ||\
    (CANARD_STM32_CAN_TIR_IDE != CANARD_STM32_CAN_RIR_IDE)
# error "RIR bits do not match TIR bits, TIR --> libcanard conversion is not possible"
#endif

    uint32_t out = 0;

    if ((id & CANARD_STM32_CAN_RIR_IDE) == 0)
    {
        out = (CANARD_CAN_STD_ID_MASK & (id >> 21));
    }
    else
    {
        out = (CANARD_CAN_EXT_ID_MASK & (id >> 3)) | CANARD_CAN_FRAME_EFF;
    }

    if ((id & CANARD_STM32_CAN_RIR_RTR) != 0)
    {
        out |= CANARD_CAN_FRAME_RTR;
    }

    return out;
}


static bool waitMSRINAKBitStateChange(volatile const CanardSTM32CANType* const bxcan, const bool target_state)
{
    /**
     * A properly functioning bus will exhibit 11 consecutive recessive bits at the end of every correct transmission,
     * or while the bus is idle. The 11 consecutive recessive bits are made up of:
     *  1 bit - acknowledgement delimiter
     *  7 bit - end of frame bits
     *  3 bit - inter frame space
     * This adds up to 11; therefore, it is not really necessary to wait longer than a few frame TX intervals.
     */
    static const unsigned TimeoutMilliseconds = 1000;

    for (unsigned wait_ack = 0; wait_ack < TimeoutMilliseconds; wait_ack++)
    {
        const bool state = (bxcan->MSR & CANARD_STM32_CAN_MSR_INAK) != 0;
        if (state == target_state)
        {
            return true;
        }
        
//        for(uint16_t i= 0; i< 0xFFF0;i++){}

        // Sleep 1 millisecond
        usleep(1000);           // TODO: This function may be missing on some platforms
    }

    return false;
}


static void processErrorStatus(PortCAN_t PortCAN)
{
    /*
     * Aborting TX transmissions if abort on error was requested
     * Updating error counter
     */
	volatile CanardSTM32CANType* const BXCAN = bxCAN[PortCAN];
    const uint8_t lec = (uint8_t)((BXCAN->ESR & CANARD_STM32_CAN_ESR_LEC_MASK) >> CANARD_STM32_CAN_ESR_LEC_SHIFT);

    if (lec != 0)
    {
        BXCAN->ESR = 0;                 // This action does only affect the LEC bits, other bits are read only!
        g_stats[PortCAN].error_count++;

        // Abort pending transmissions if auto abort on error is enabled, or if we're in bus off mode
        if (g_abort_tx_on_error || (BXCAN->ESR & CANARD_STM32_CAN_ESR_BOFF))
        {
            BXCAN->TSR = CANARD_STM32_CAN_TSR_ABRQ0 | CANARD_STM32_CAN_TSR_ABRQ1 | CANARD_STM32_CAN_TSR_ABRQ2;
        }
    }
}

#if USERXINT
static void nvicEnableVector(IRQn_Type irq)
{
  NVIC_SetPriority(irq, NVIC_EncodePriority(NVIC_GetPriorityGrouping(),0, 0));
  NVIC_EnableIRQ(irq);
}
#endif

int canardSTM32Init(const CanardSTM32CANTimings* const timings,
                    const CanardSTM32IfaceMode iface_mode,	PortCAN_t PortCAN
)
{
    /*
     * Paranoia time.
     */
	static bool PortInit[PortCANLast] = {0};
	PortInit[PortCAN] = true;
#if USERXINT
		VarInitQ(&CANRxQ[PortCAN],CANRxQ[PortCAN].Len);
#endif
		volatile CanardSTM32CANType* BXCAN = bxCAN[PortCAN];
		BXCAN->MCR = CANARD_STM32_CAN_MCR_RESET;

    if ((iface_mode != CanardSTM32IfaceModeNormal) &&
        (iface_mode != CanardSTM32IfaceModeSilent) &&
        (iface_mode != CanardSTM32IfaceModeAutomaticTxAbortOnError))
    {
        return -CANARD_ERROR_INVALID_ARGUMENT;
    }

    if ((timings == NULL) ||
        (timings->bit_rate_prescaler < 1) || (timings->bit_rate_prescaler > 1024) ||
        (timings->max_resynchronization_jump_width < 1) || (timings->max_resynchronization_jump_width > 4) ||
        (timings->bit_segment_1 < 1) || (timings->bit_segment_1 > 16) ||
        (timings->bit_segment_2 < 1) || (timings->bit_segment_2 > 8))
    {
  		char *str = "invalid arg\n";
  		SerialTxQWait(str,strlen(str));
        return -CANARD_ERROR_INVALID_ARGUMENT;
    }

    /*
     * Initial setup
     */
    memset(&g_stats[PortCAN], 0, sizeof(g_stats[0]));

    g_abort_tx_on_error = (iface_mode == CanardSTM32IfaceModeAutomaticTxAbortOnError);
    if((PortCAN2 == PortCAN)&&(!PortInit[PortCAN1]))				// If we're using CAN2, we MUST initialize CAN1 first, because CAN2 is a slave to CAN1
    {
			CANARD_STM32_CAN1->IER = 0;                                 // We need no interrupts
			CANARD_STM32_CAN1->MCR &= ~CANARD_STM32_CAN_MCR_SLEEP;      // Exit sleep mode
			CANARD_STM32_CAN1->MCR |= CANARD_STM32_CAN_MCR_INRQ;        // Request init

			if (!waitMSRINAKBitStateChange(CANARD_STM32_CAN1, true))    // Wait for synchronization
			{
				CANARD_STM32_CAN1->MCR = CANARD_STM32_CAN_MCR_RESET;
	  		char *str = "1inak not set\n";
	  		SerialTxQWait(str,strlen(str));
				return -CANARD_STM32_ERROR_MSR_INAK_NOT_SET;
			}
			// CAN1 will be left in the initialization mode forever, in this mode it does not affect the bus at all.
    }
    BXCAN->IER = 0;                                             // We need no interrupts
    BXCAN->MCR &= ~CANARD_STM32_CAN_MCR_SLEEP;                  // Exit sleep mode
    BXCAN->MCR |= CANARD_STM32_CAN_MCR_INRQ;                    // Request init

    if (!waitMSRINAKBitStateChange(BXCAN, true))                // Wait for synchronization
    {
  		char *str = "2inak not set2\n";
  		SerialTxQWait(str,strlen(str));
        BXCAN->MCR = CANARD_STM32_CAN_MCR_RESET;
        return -CANARD_STM32_ERROR_MSR_INAK_NOT_SET;
    }

    /*
     * Hardware initialization (the hardware has already confirmed initialization mode, see above)
     */
    BXCAN->MCR = CANARD_STM32_CAN_MCR_ABOM | CANARD_STM32_CAN_MCR_AWUM | CANARD_STM32_CAN_MCR_INRQ;  // RM page 648

    BXCAN->BTR = (((timings->max_resynchronization_jump_width - 1) &    3U) << 24) |
                 (((timings->bit_segment_1 - 1)                    &   15U) << 16) |
                 (((timings->bit_segment_2 - 1)                    &    7U) << 20) |
                 ((timings->bit_rate_prescaler - 1)                & 1023U) |
                 ((iface_mode == CanardSTM32IfaceModeSilent) ? CANARD_STM32_CAN_BTR_SILM : 0);
    CANARD_ASSERT(0 == BXCAN->IER);             // Making sure the iterrupts are indeed disabled
#if USERXINT
    BXCAN->IER = CANARD_STM32_CAN_IER_FMPIE0|CANARD_STM32_CAN_IER_FMPIE1;//|CANARD_STM32_CAN_IER_TMEIE;
#endif
    BXCAN->MCR &= ~CANARD_STM32_CAN_MCR_INRQ;   // Leave init mode
    if (!waitMSRINAKBitStateChange(BXCAN, false))
    {
  		char *str = "3inak not cleared\n";
  		SerialTxQWait(str,strlen(str));
        BXCAN->MCR = CANARD_STM32_CAN_MCR_RESET;
        return -CANARD_STM32_ERROR_MSR_INAK_NOT_CLEARED;
    }

    /*
     * Default filter configuration. Note that ALL filters are available ONLY via CAN1! NO! CAN 3 has it's own
     * CAN2 filters are offset by 14.
     * We use 14 filters at most always which simplifies the code and ensures compatibility with all
     * MCU within the STM32 family.
     */
    if(((PortCAN2 == PortCAN)&&(!PortInit[PortCAN1]))||((PortCAN1 == PortCAN)))
    {
        uint32_t fmr = CANARD_STM32_CAN1->FMR & 0xFFFFC0F1;
        fmr |= CANARD_STM32_NUM_ACCEPTANCE_FILTERS << 8;                // CAN2 start bank = 14 (if CAN2 is present)
        CANARD_STM32_CAN1->FMR = fmr | CANARD_STM32_CAN_FMR_FINIT;
        CANARD_ASSERT(((CANARD_STM32_CAN1->FMR >> 8) & 0x3F) == CANARD_STM32_NUM_ACCEPTANCE_FILTERS);
        CANARD_STM32_CAN1->FM1R = 0;                                        // Indentifier Mask mode
        CANARD_STM32_CAN1->FS1R = 0x0FFFFFFF;                               // All 32-bit
        // Filters are alternating between FIFO0 and FIFO1 in order to equalize the load.
        // This will cause occasional priority inversion and frame reordering on reception,
        // but that is acceptable for UAVCAN, and a majority of other protocols will tolerate
        // this too, since there will be no reordering within the same CAN ID.
        CANARD_STM32_CAN1->FFA1R = 0x0AAAAAAA;
    }
    if(PortCAN1 == PortCAN)
    {
    	BXCAN->FilterRegister[0].FR1 = 0;
    	BXCAN->FilterRegister[0].FR2 = 0;
    	BXCAN->FA1R = 0;
    	BXCAN->FA1R |= 1;                                      	// One filter enabled
    	BXCAN->FMR &= ~CANARD_STM32_CAN_FMR_FINIT;              // Leave initialization mode
    }
    else if(PortCAN2 == PortCAN)
    {
        CANARD_STM32_CAN1->FilterRegister[CANARD_STM32_NUM_ACCEPTANCE_FILTERS].FR1 = 0;
        CANARD_STM32_CAN1->FilterRegister[CANARD_STM32_NUM_ACCEPTANCE_FILTERS].FR2 = 0;
        if(!PortInit[PortCAN1])
        {BXCAN->FA1R = 0;}
        else
        {CANARD_STM32_CAN1->FA1R &= ~(((1<<(2*CANARD_STM32_NUM_ACCEPTANCE_FILTERS))-1)&~((1<<CANARD_STM32_NUM_ACCEPTANCE_FILTERS)-1));}	// clear all bits up to 14 filters
        CANARD_STM32_CAN1->FA1R |= (1 << CANARD_STM32_NUM_ACCEPTANCE_FILTERS);  // One filter enabled
        CANARD_STM32_CAN1->FMR &= ~CANARD_STM32_CAN_FMR_FINIT;              // Leave initialization mode
    }
    else if(PortCAN3 == PortCAN)
	{
		BXCAN->FilterRegister[0].FR1 = 0;
		BXCAN->FilterRegister[0].FR2 = 0;
		BXCAN->FA1R = 0;
		BXCAN->FA1R |= 1;                                        		// One filter enabled
		BXCAN->FMR &= ~CANARD_STM32_CAN_FMR_FINIT;              // Leave initialization mode
	}
#if USERXINT
		IRQn_Type RX0irq,RX1irq;//,TXirq;
		RX0irq = (PortCAN1==PortCAN)?(CAN1_RX0_IRQn):((PortCAN2==PortCAN)?(CAN2_RX0_IRQn):(CAN3_RX0_IRQn));
		RX1irq = (PortCAN1==PortCAN)?(CAN1_RX1_IRQn):((PortCAN2==PortCAN)?(CAN2_RX1_IRQn):(CAN3_RX1_IRQn));
		//TXirq  = (PortCAN1==PortCAN)?(CAN1_TX_IRQn ):((PortCAN2==PortCAN)?(CAN2_TX_IRQn ):(CAN3_TX_IRQn ));
		//nvicEnableVector(TXirq);
		nvicEnableVector(RX0irq);
		nvicEnableVector(RX1irq);
#endif
		char *str = "4canard stm init ok\n";
		SerialTxQWait(str,strlen(str));
    return 0;
}


int canardSTM32Transmit(const CanardCANFrame* const frame,PortCAN_t PortCAN)
{
	volatile CanardSTM32CANType* BXCAN = bxCAN[PortCAN];
	if (frame == NULL)
    {
        return -CANARD_ERROR_INVALID_ARGUMENT;
    }

    if (frame->id & CANARD_CAN_FRAME_ERR)
    {
        return -CANARD_STM32_ERROR_UNSUPPORTED_FRAME_FORMAT;
    }

    /*
     * Handling error status might free up some slots through aborts
     */
    processErrorStatus(PortCAN);

    /*
     * Seeking an empty slot, checking if priority inversion would occur if we enqueued now.
     * We can always enqueue safely if all TX mailboxes are free and no transmissions are pending.
     */
    uint8_t tx_mailbox = 0xFF;

    static const uint32_t AllTME = CANARD_STM32_CAN_TSR_TME0 | CANARD_STM32_CAN_TSR_TME1 | CANARD_STM32_CAN_TSR_TME2;

    if ((BXCAN->TSR & AllTME) != AllTME)                // At least one TX mailbox is used, detailed check is needed
    {
        const bool tme[3] =
        {
            (BXCAN->TSR & CANARD_STM32_CAN_TSR_TME0) != 0,
            (BXCAN->TSR & CANARD_STM32_CAN_TSR_TME1) != 0,
            (BXCAN->TSR & CANARD_STM32_CAN_TSR_TME2) != 0
        };

        for (uint8_t i = 0; i < 3; i++)
        {
            if (tme[i])                                 // This TX mailbox is free, we can use it
            {
                tx_mailbox = i;
            }
            else                                        // This TX mailbox is pending, check for priority inversion
            {
                if (!isFramePriorityHigher(frame->id, convertFrameIDRegisterToCanard(BXCAN->TxMailbox[i].TIR)))
                {
                    // There's a mailbox whose priority is higher or equal the priority of the new frame.
                    return 0;                           // Priority inversion would occur! Reject transmission.
                }
            }
        }

        if (tx_mailbox == 0xFF)
        {
            /*
             * All TX mailboxes are busy (this is highly unlikely); at the same time we know that there is no
             * higher or equal priority frame that is currently pending. Therefore, priority inversion has
             * just happend (sic!), because we can't enqueue the higher priority frame due to all TX mailboxes
             * being busy. This scenario is extremely unlikely, because in order for it to happen, the application
             * would need to transmit 4 (four) or more CAN frames with different CAN ID ordered from high ID to
             * low ID nearly at the same time. For example:
             *  1. 0x123        <-- Takes mailbox 0 (or any other)
             *  2. 0x122        <-- Takes mailbox 2 (or any other)
             *  3. 0x121        <-- Takes mailbox 1 (or any other)
             *  4. 0x120        <-- INNER PRIORITY INVERSION HERE (only if all three TX mailboxes are still busy)
             * This situation is even less likely to cause any noticeable disruptions on the CAN bus. Despite that,
             * it is better to warn the developer about that during debugging, so we fire an assertion failure here.
             * It is perfectly safe to remove it.
             */
#if CANARD_STM32_DEBUG_INNER_PRIORITY_INVERSION
            CANARD_ASSERT(!"CAN PRIO INV");
#endif
            return 0;
        }
    }
    else                                                // All TX mailboxes are free, use first
    {
        tx_mailbox = 0;
    }

    CANARD_ASSERT(tx_mailbox < 3);                      // Index check - the value must be correct here

    /*
     * By this time we've proved that a priority inversion would not occur, and we've also found a free TX mailbox.
     * Therefore it is safe to enqueue the frame now.
     */
    volatile CanardSTM32TxMailboxType* const mb = &BXCAN->TxMailbox[tx_mailbox];

    mb->TDTR = frame->data_len;                         // DLC equals data length except in CAN FD

    mb->TDHR = (((uint32_t)frame->data[7]) << 24) |
               (((uint32_t)frame->data[6]) << 16) |
               (((uint32_t)frame->data[5]) <<  8) |
               (((uint32_t)frame->data[4]) <<  0);
    mb->TDLR = (((uint32_t)frame->data[3]) << 24) |
               (((uint32_t)frame->data[2]) << 16) |
               (((uint32_t)frame->data[1]) <<  8) |
               (((uint32_t)frame->data[0]) <<  0);

    mb->TIR = convertFrameIDCanardToRegister(frame->id) | CANARD_STM32_CAN_TIR_TXRQ;    // Go.

    /*
     * The frame is now enqueued and pending transmission.
     */
    return 1;
}


int canardSTM32Receive(CanardCANFrame* const out_frame,PortCAN_t PortCAN)
{
	if ((out_frame == NULL)||(PortCAN >= PortCANLast)||(PortCAN < 0))
	{
			return -CANARD_ERROR_INVALID_ARGUMENT;
	}
#if USERXINT
	uint16_t Len = sizeof(CanardCANFrame),qLen;
	processErrorStatus(PortCAN);
	if((qLen=QLen(&CANRxQ[PortCAN])) >= Len)																			// enough data to process at least one frame
	{
		if(DeQ(&CANRxQ[PortCAN],(uint8_t*)out_frame,&Len))
		{
			return(1);
		}
	}
	else if(qLen)																																	// something wrong >0 and less than frame bytes
	{
		g_stats[PortCAN].error_count ++;
		DeQ(&CANRxQ[PortCAN],NULL,&Len);																						// clear the RX Que
	}

	return(0);
#else
	volatile CanardSTM32CANType* const BXCAN = bxCAN[PortCAN];
    volatile uint32_t* RFxR[2] =
    {
        &BXCAN->RF0R,
        &BXCAN->RF1R
    };

    /*
     * This function must be polled periodically, so we use this opportunity to do it.
     */
    processErrorStatus(PortCAN);

    /*
     * Reading the TX FIFO
     */
    for (uint_fast8_t i = 0; i < 2; i++)
    {
        volatile CanardSTM32RxMailboxType* const mb = &BXCAN->RxMailbox[i];

        if (((*RFxR[i]) & CANARD_STM32_CAN_RFR_FMP_MASK) != 0)
        {
            if (*RFxR[i] & CANARD_STM32_CAN_RFR_FOVR)
            {
                g_stats[PortCAN].rx_overflow_count++;
            }

            out_frame->id = convertFrameIDRegisterToCanard(mb->RIR);

            out_frame->data_len = (uint8_t)(mb->RDTR & CANARD_STM32_CAN_RDTR_DLC_MASK);

            // Caching to regular (non volatile) memory for faster reads
            const uint32_t rdlr = mb->RDLR;
            const uint32_t rdhr = mb->RDHR;

            out_frame->data[0] = (uint8_t)(0xFF & (rdlr >>  0));
            out_frame->data[1] = (uint8_t)(0xFF & (rdlr >>  8));
            out_frame->data[2] = (uint8_t)(0xFF & (rdlr >> 16));
            out_frame->data[3] = (uint8_t)(0xFF & (rdlr >> 24));
            out_frame->data[4] = (uint8_t)(0xFF & (rdhr >>  0));
            out_frame->data[5] = (uint8_t)(0xFF & (rdhr >>  8));
            out_frame->data[6] = (uint8_t)(0xFF & (rdhr >> 16));
            out_frame->data[7] = (uint8_t)(0xFF & (rdhr >> 24));

            // Release FIFO entry we just read
            *RFxR[i] = CANARD_STM32_CAN_RFR_RFOM | CANARD_STM32_CAN_RFR_FOVR | CANARD_STM32_CAN_RFR_FULL;

            // Reading successful
            return 1;
        }
    }

    // No frames to read
    return 0;
#endif
}


int canardSTM32ConfigureAcceptanceFilters(const CanardSTM32AcceptanceFilterConfiguration* const filter_configs,
                                          const unsigned num_filter_configs,PortCAN_t PortCAN)
{
	volatile CanardSTM32CANType* BXCAN = bxCAN[PortCAN];

    // Note that num_filter_configs = 0 is a valid configuration, which rejects all frames.
    if ((filter_configs == NULL) ||
        (num_filter_configs > CANARD_STM32_NUM_ACCEPTANCE_FILTERS))
    {
        return -CANARD_ERROR_INVALID_ARGUMENT;
    }

    /*
     * First we disable all filters. This may cause momentary RX frame losses, but the application
     * should be able to tolerate that.
     */
    if((PortCAN1 == PortCAN)||(PortCAN3 == PortCAN))
    {
       	BXCAN->FA1R &= ~((1<<CANARD_STM32_NUM_ACCEPTANCE_FILTERS)-1);	// clear all bits up to 14 filters
    }
    else if(PortCAN2 == PortCAN)
    {
        CANARD_STM32_CAN1->FA1R &= ~(((1<<(2*CANARD_STM32_NUM_ACCEPTANCE_FILTERS))-1)&~((1<<CANARD_STM32_NUM_ACCEPTANCE_FILTERS)-1));	// clear all bits up to 14 filters
       // CANARD_STM32_CAN1->FA1R = 0;
    }


    /*
     * Having filters disabled we can update the configuration.
     * Register mapping: FR1 - ID, FR2 - Mask
     */
    for (unsigned i = 0; i < num_filter_configs; i++)
    {
        /*
         * Converting the ID and the Mask into the representation that can be chewed by the hardware.
         * If Mask asks us to accept both STDID and EXTID, we need to use EXT mode on the filter,
         * otherwise it will reject all EXTID frames. This logic is not documented in the RM.
         * True story. This is sacred knowledge! Don't tell the unworthy.
         */
        uint32_t id   = 0;
        uint32_t mask = 0;

        const CanardSTM32AcceptanceFilterConfiguration* const cfg = filter_configs + i;

        if ((cfg->id & CANARD_CAN_FRAME_EFF) || !(cfg->mask & CANARD_CAN_FRAME_EFF))
        {
            // The application wants to trick us by requesting both ext and std frames!
            id   = (cfg->id   & CANARD_CAN_EXT_ID_MASK) << 3;   // Little they know, we can handle that correctly.
            mask = (cfg->mask & CANARD_CAN_EXT_ID_MASK) << 3;   // See?
            id |= CANARD_STM32_CAN_RIR_IDE;                     // That's how we do this. Where's my sunglasses?
        }
        else
        {
            id   = (cfg->id   & CANARD_CAN_STD_ID_MASK) << 21;  // Regular std frames, nothing fancy.
            mask = (cfg->mask & CANARD_CAN_STD_ID_MASK) << 21;  // Boring.
        }

        if (cfg->id & CANARD_CAN_FRAME_RTR)
        {
            id |= CANARD_STM32_CAN_RIR_RTR;
        }

        if (cfg->mask & CANARD_CAN_FRAME_EFF)
        {
            mask |= CANARD_STM32_CAN_RIR_IDE;
        }

        if (cfg->mask & CANARD_CAN_FRAME_RTR)
        {
            mask |= CANARD_STM32_CAN_RIR_RTR;
        }

        /*
         * Applying the converted representation to the registers.
         */
        unsigned filter_index;
        if((PortCAN1 == PortCAN)||(PortCAN3 == PortCAN))
        {
        	filter_index = i;
            BXCAN->FilterRegister[filter_index].FR1 = id;
            BXCAN->FilterRegister[filter_index].FR2 = mask;
            BXCAN->FA1R |= 1U << filter_index;      // Enable
        }
        else if(PortCAN2 == PortCAN)
        {
        	filter_index = i + CANARD_STM32_NUM_ACCEPTANCE_FILTERS;
        	CANARD_STM32_CAN1->FilterRegister[filter_index].FR1 = id;
        	CANARD_STM32_CAN1->FilterRegister[filter_index].FR2 = mask;
            CANARD_STM32_CAN1->FA1R |= 1U << filter_index;      // Enable
        }
    }

    return 0;
}


CanardSTM32Stats canardSTM32GetStats(PortCAN_t PortCAN)
{
    return g_stats[PortCAN];
}
#if USERXINT
inline void handleRxInterrupt(PortCAN_t PortCAN,uint8_t fifo_index, volatile CanardSTM32CANType* const can_)
{
	 	 volatile uint32_t* const rfr_reg = (fifo_index == 0) ? &can_->RF0R : &can_->RF1R;
#if 0
    if ((*rfr_reg & CANARD_STM32_CAN_RFR_FMP_MASK) == 0)
    {
        UAVCAN_ASSERT(0);  // Weird, IRQ is here but no data to read
        return;
    }
#endif
    if ((*rfr_reg & CANARD_STM32_CAN_RFR_FOVR) != 0)	//Register overflow as a hardware error
    {
      g_stats[PortCAN].rx_overflow_count++;
    }
    // Read the frame contents
    CanardCANFrame frame;
    volatile CanardSTM32RxMailboxType* rf = &can_->RxMailbox[fifo_index];
    frame.id  = 0;
    if ((rf->RIR & CANARD_STM32_CAN_RIR_IDE) == 0)
    {
        frame.id = CANARD_CAN_STD_ID_MASK & (rf->RIR >> 21);
    }
    else
    {
    	frame.id = (CANARD_CAN_EXT_ID_MASK & (rf->RIR >> 3))| CANARD_CAN_FRAME_EFF;
    }
    if ((rf->RIR & CANARD_STM32_CAN_RIR_RTR) != 0)
    {
        frame.id |= CANARD_CAN_FRAME_RTR;
    }

    frame.data_len = (uint8_t)(rf->RDTR & CANARD_STM32_CAN_RDTR_DLC_MASK);

    frame.data[0] = (uint8_t)(0xFF & (rf->RDLR >> 0));
    frame.data[1] = (uint8_t)(0xFF & (rf->RDLR >> 8));
    frame.data[2] = (uint8_t)(0xFF & (rf->RDLR >> 16));
    frame.data[3] = (uint8_t)(0xFF & (rf->RDLR >> 24));
    frame.data[4] = (uint8_t)(0xFF & (rf->RDHR >> 0));
    frame.data[5] = (uint8_t)(0xFF & (rf->RDHR >> 8));
    frame.data[6] = (uint8_t)(0xFF & (rf->RDHR >> 16));
    frame.data[7] = (uint8_t)(0xFF & (rf->RDHR >> 24));

    *rfr_reg = CANARD_STM32_CAN_RFR_RFOM | CANARD_STM32_CAN_RFR_FOVR | CANARD_STM32_CAN_RFR_FULL;  // Release FIFO entry we just read

    if(!EnQ(&CANRxQ[PortCAN],(uint8_t*)&frame,sizeof(frame)))											//Store into the FIFO buffer
    {
      g_stats[PortCAN].rx_overflow_count += sizeof(frame);
    }
}

void CAN1_RX0_IRQHandler(void)
{
	handleRxInterrupt(0,0, CANARD_STM32_CAN1);
}
void CAN1_RX1_IRQHandler(void)
{
	handleRxInterrupt(0,1, CANARD_STM32_CAN1);
}
void CAN2_RX0_IRQHandler(void)
{
	handleRxInterrupt(1,0, CANARD_STM32_CAN2);
}
void CAN2_RX1_IRQHandler(void)
{
	handleRxInterrupt(1,1, CANARD_STM32_CAN2);
}
void CAN3_RX0_IRQHandler(void)
{
	handleRxInterrupt(2,0, CANARD_STM32_CAN3);
}
void CAN3_RX1_IRQHandler(void)
{
	handleRxInterrupt(2,1, CANARD_STM32_CAN3);
}
#endif
#if 0
inline void handleTxInterrupt(volatile CanardSTM32CANType* const can_)
{
    // TXOK == false means that there was a hardware failure
    if (can_->TSR  & CANARD_STM32_CAN_TSR_RQCP0)// request completed mailbox 0
    {
        const bool txok = can_->TSR & CANARD_STM32_CAN_TSR_TXOK0;
        can_->TSR = CANARD_STM32_CAN_TSR_RQCP0;
       // handleTxMailboxInterrupt(0, txok, utc_usec); TODO ad code to handle this
    }
    if (can_->TSR & CANARD_STM32_CAN_TSR_RQCP1)// request completed mailbox 1
    {
        const bool txok = can_->TSR & CANARD_STM32_CAN_TSR_TXOK1;
        can_->TSR = CANARD_STM32_CAN_TSR_RQCP1;
        //handleTxMailboxInterrupt(1, txok, utc_usec); TODO add code to handle this
    }
    if (can_->TSR & CANARD_STM32_CAN_TSR_RQCP2)	// request completed mailbox 2
    {
        const bool txok = can_->TSR & CANARD_STM32_CAN_TSR_TXOK2;
        can_->TSR = CANARD_STM32_CAN_TSR_RQCP2;
        // handleTxMailboxInterrupt(2, txok, utc_usec); TODO add code to handle this
    }
    //update_event_.signalFromInterrupt();

    //pollErrorFlagsFromISR();

    #if UAVCAN_STM32_FREERTOS
    //update_event_.yieldFromISR();
    #endif
}

void CAN1_TX_IRQHandler(void)
{
}
void CAN1_TX_IRQHandler (void)
{
}
void CAN3_TX_IRQHandler (void)
{
}
void CAN2_TX_IRQHandler (void)
{
}
#endif
