/*
 * PacketHandler.c
 * We must be able to handle missing chunks of data
 * and not throw out good data because of an error
 * to do this we must reprocess a buffer if there is a check sum failure
 * starting on the next byte after the first found start sequence
 * We will need to keep a large buffer for processing and only discard
 * data once a valid packet is found
 * The buffer will need to be shifted once valid data found
 * and more data added to it as avilable until full
 *  Created on: 19/03/2016
 *      Author: kentm
 */
// ******************** includes *************************************
#include <Pkt.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>

#include "PacketHandler.h"
#include "Pkt.h"
#include "serial.h"

// ******************** defines and typedefs *************************

enum {
  StateSTX1=0,
  StateSTX2,
  StateLengthMSB,
  StateLengthLSB,
  StateData,
  StateChkSumMSB,
  StateChkSumLSB,
  StateETX,
  StateLast
};

// ******************** local variables ******************************
static uint8_t LBuff[16+MAX_DATA_SIZE*2];																				// fit two full packets
static uint16_t LBuffIdx=0;
static uint16_t LBuffSize=0;
static uint16_t CurrCRC;
// ******************** local function prototypes ********************
static bool ProcessData(uint8_t *Src,uint16_t SrcLen,uint16_t *SrcIdx,uint8_t *OutData,uint16_t *OutLen,bool *res);
static bool ParseData(uint8_t *Src, uint16_t *SrcIdx,uint16_t SrcLen,uint8_t *OutData,uint16_t *OutLen, bool *error);
static uint16_t calc_CRC(unsigned char *p, uint16_t packetLength);
static void update_CRC(uint8_t *Data, uint16_t len);
// ********************* Implementation ******************************
void PacketHandler_Init(void)
{
	InitSerial();
}

bool StartPacket(uint16_t len)
{
	Header_t Head;
	Head.LenHi = (len>>8)&0xff;
	Head.LenLo = len&0xff;
	Head.STXHi = (START>>8)&0xff;
	Head.STXLo = START&0xff;
	CurrCRC = 0;
	return(sizeof(Head) == SerialTxQ((uint8_t*)&Head,sizeof(Head)));
}

bool AppendPacket(uint8_t *Data,uint16_t len)
{
	update_CRC(Data,len);
	return(len == SerialTxQ(Data,len));
}

bool SendPacket(void)
{
	Tail_t Tail;
	Tail.CRCHi = (CurrCRC>>8)&0xff;
	Tail.CRCLo = CurrCRC&0xff;
	Tail.etx = ETX;
	return(sizeof(Tail) == SerialTxQ((uint8_t*)&Tail,sizeof(Tail)));
}

// read serial port for data stream and copy data into buffer
bool Packet_Poll(uint8_t *outData,uint16_t *outLen)
{
	bool res;

	if(LBuffSize < sizeof(LBuff))																									// we can add some data
	{
		uint16_t len;
		if((len=SerialRxQSize()))
		{
			uint16_t space = sizeof(LBuff)-LBuffSize;
			if(len > space ){len=space;}
			if(SerialRxQBuff(&LBuff[LBuffSize],&len))
			{
				LBuffSize += len;
			}
		}
	}

	// now the buffer has been topped up process data
	if((LBuffIdx >=sizeof(LBuff))||																								// if hit end of buffer or
		 (ProcessData(LBuff,LBuffSize,&LBuffIdx,outData,outLen,&res)))						  // if some data needs to be removed
	{
		LBuffSize -= LBuffIdx;
		if(LBuffSize > 0)																														// if any data left
		{
			memmove(LBuff,&LBuff[LBuffIdx],LBuffSize);																// move data to start
		}
		LBuffIdx = 0;																																// set index to start
	}
	return(res);																																	// return if valid data was found
}

static bool ProcessData(uint8_t *Src,uint16_t SrcLen,uint16_t *SrcIdx,uint8_t *OutData,uint16_t *OutLen,bool *res)
{
	bool error = false;
	*res = false;
	while((false == error)&&(false == *res)&&(*SrcIdx < SrcLen))
	{
		*res = ParseData(Src,SrcIdx,SrcLen,OutData,OutLen,&error);
	}
	return(*res||error);
}


static bool ParseData(uint8_t *Src, uint16_t *SrcIdx,uint16_t SrcLen, uint8_t *OutData,uint16_t *OutLen, bool *error)
{
	static uint8_t State=StateSTX1,ch;
  static uint16_t count,Len,Sum,Start,DataStart;
  bool Res = false;
  ch = Src[*SrcIdx];
  switch (State)
  {
  	case   StateSTX1      :	Start = *SrcIdx;
  													if(ch == START >> 8  )															// if match of first byte
  													{
    													if(0 == *SrcIdx)																	// if we are at 0 when first id found
    													{
    														State++;																				// go to next state
    													}
    													else																							// else not a 0, so move buffer
    													{
      													*error = true;																	// force move of buffer back to zero
      													Start--;																				// idx will be set to start then 1 added at end
    													}
  													}
  													break;
  	case   StateSTX2      : State = (ch == (START&0xff))?(State+1):(StateSTX1); break;
    case   StateLengthMSB : State = (ch <= ((MAX_DATA_SIZE+2)>>8))?(State+1):(StateSTX1);
                            Len = ((uint16_t)ch)<<8;DataStart=(*SrcIdx)+2;			break;
    case   StateLengthLSB : Len += ch;
    												State = ((Len <= (MAX_DATA_SIZE))&&(0!=Len))?(State+1):(StateSTX1);
														                                        						break;
    case   StateData      :	count  = (SrcLen-DataStart);												// how many bytes avialable in buffer?
    												if(count > Len){count = Len;}												// if more than block, set to block size
    												(*SrcIdx) = DataStart+(count-1);										// set to position 1 less than count
    												State = (count >= Len)?(State+1):(State);    				break;
    case   StateChkSumMSB : Sum = ((uint16_t)ch)<<8;State++;										break;
    case   StateChkSumLSB : Sum += ch;
    												if(Sum == calc_CRC(&Src[Start+4] ,Len))
    												{
    													State = StateETX;
    												}
    												else
    												{
    													*error = true;
    													State = StateSTX1;
    												}
    																																						break;
    case   StateETX :				if(Len >*OutLen)
    												{
    													*error = true;
    												}
    												else
    												{
    													Res = true;
    													memcpy(OutData,&Src[DataStart],Len);
    													*OutLen = Len;
    												}
    												Start = *SrcIdx;																	// will get set at end
    												State = StateSTX1;																	break;
  }
  if(StateSTX1 == State)
  {
  	*SrcIdx = Start;
  }
  (*SrcIdx)++;
  return(Res);
}

static void update_CRC(uint8_t *Data, uint16_t len)
{
  static uint16_t i, j;
  for (i = 0; i < len; i++)
  {
      CurrCRC = CurrCRC ^ Data[i] << 8;
      for(j=0;j < 8; j++)
      {
          if (CurrCRC & 0x8000)
          	CurrCRC = CurrCRC << 1 ^ 0x1021;
          else
          	CurrCRC = CurrCRC << 1;
      }
  }
}

static uint16_t calc_CRC(unsigned char *p, uint16_t packetLength)
{
    uint16_t crc = 0;
    static uint16_t i, j;
    for (i = 0; i < packetLength; i++)
    {
        crc = crc ^ p[i] << 8;
        for(j=0;j < 8; j++)
        {
            if (crc & 0x8000)
                crc = crc << 1 ^ 0x1021;
            else
                crc = crc << 1;
        }
    }
    return crc;
}

/** end of PacketHandler.c *******************************************/


