/*
 * main_canard.c
 *
 *  Created on: 22/07/2019
 *      Author: kentm
 */

#include <stdio.h>
#include <stdint.h>
#include "canard.h"
#include "canard_stm32.h"
#include "FreeRTOS.h"
#include "task.h"
#include "cmsis_os.h"
#include "VisibleFlightController.h"
#include "UseFlightController.h"
#include "IgnoreCANPort.h"
#include "Allocation.h"
#include "GetNodeInfo.h"
#include "NodeStatus.h"
#include "RawCommand.h"
#include "GetSet.h"
#include "ArrayCommand.h"
#include "Command.h"
#include "PWM.h"
#include "CAN.h"
#include "gpio.h"
#include "NVM.h"
#include "dynNodeIDAlloc.h"
#include "System_Config.h"

// ******************** defines and typedefs ***********************************
#if (APP_CURR == APP_BRIDGE)
#define RFDCANLINK 1
#include "PacketHandler.h"
#include "FlightControllerHealth.h"
#elif (APP_CURR == APP_ARBITER)
#define RFDARBITER 1
#include "FlightControllerHealth.h"
#include "Arbiter.h"
#include "ArbiterTest.h"
#elif (APP_CURR == APP_ENDP)
#define RFDENDPOINT 1
#include "Voters.h"
#endif

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof(arr[0]))
#define LIMIT(a,min,max)(((a)<(min))?(min):(((a)>(max))?(max):(a)))
#define MIN(a,b) ((a<b)?(a):(b))
#define MAX(a,b) ((a>b)?(a):(b))

#define ESC_RAWCOMMAND_MAX_VALUE                   8182
#define ESC_RAWCOMMAND_MIN_VALUE                   160
#define ESC_RAWCOMMAND_RANGE											(ESC_RAWCOMMAND_MAX_VALUE-ESC_RAWCOMMAND_MIN_VALUE)

#if RFDARBITER
#define DEF_NODEID 					 100
#define USEFC_PUB_MS		     100																								// how long between publishing use FC messages
#define HEALTH_UPDATE_MS		 50																								// how long between health checks

#elif RFDENDPOINT
#define DEF_NODEID 					 110
#define CHECKFC_PERIOD_mS		 50																								// interval to update Flight controller list
#define VIS_CHECK_MS	 50																											// how long between checking visibility matrix
#define VS_TIMEOUT_MS	 100																										  // how long after no servo messages visibility is invalidated
#define VIS_PUB_MS		 100																											// how long between publishing visibility messages

#elif RFDCANLINK
#define DEF_NODEID 					 90
typedef struct{
	uint8_t dtype_id;													// data type id
	uint8_t transfer_id;											// transfer id variable
	int8_t	NodeID;														// nodeID
	uint32_t TickmS;													// last time of activity
} tidEntry_t;
#define TID_TIMEOUT_MS	5000
#define TID_CHECK_MS		1000
typedef struct
{
    uint16_t data_type_id;                  ///< 0 to 255 for services, 0 to 65535 for messages
    uint8_t transfer_type;                  ///< See CanardTransferType
    uint8_t transfer_id;                    ///< 0 to 31
    uint8_t priority;                       ///< 0 to 31
    uint8_t source_node_id;                 ///< 1 to 127, or 0 if the source is anonymous
    uint8_t dest_node_id;                 	///< 1 to 127, or 0 if the dest is anonymous
    uint64_t data_type_signature;						///< 1 -  64 bit message signature
}CANRxTransfer;
#endif
#define CANARD_SPIN_PERIOD   1000																								// node status update period
#define CANStrt 0
#define CANEnd	3
typedef struct {
	TIM_TypeDef * timer;
	uint8_t channel;
	const char * name;
} rcpwm_t;

typedef struct {
	uint8_t* name;
	int64_t val;
	int64_t min;
	int64_t max;
	int64_t defval;
} param_t;

typedef struct{
	uint16_t data_type_id;
	uint64_t out_data_type_signature;
	int8_t transfer_type;
}Msg_Filt_t;
// ******************** local variables ****************************************
osThreadId MainTaskHandle;
static uint8_t 	gNodeID[PortCANLast];
static CanardInstance g_canard[PortCANLast];              											//The library instance
static uint8_t g_canard_memory_pool[PortCANLast][3096]; 												//Arena for memory allocation, used by the library
static const CAN_Init_t CAN_Init[PortCANLast] = CAN_INIT;
#if RFDENDPOINT
static int8_t ActiveFCPort = -1;
static bool PixVisibility[PortCANLast] = { false };
static uint32_t RxFCTick[PortCANLast]={0};
#endif
#define NUM_PWM 6

typedef enum{
	Var_EEFormat=0,																																// value to define eeprom variables stored, clear if not matching
	Var_NETID0,
	Var_NETID1,
	Var_NETID2,
	Var_CAN2PWM_CHStart,
	Var_CAN2PWM_ACTStart=Var_CAN2PWM_CHStart+NUM_PWM,
	Var_CAN_TotBlocks= Var_CAN2PWM_ACTStart+NUM_PWM,
	Var_CAN_CurrBlocks=Var_CAN_TotBlocks+PortCANLast,
	Var_CAN_PeakBlocks=Var_CAN_CurrBlocks+PortCANLast,
	Var_CAN_RXOVF	=Var_CAN_PeakBlocks+PortCANLast,
	Var_CAN_RXERR = Var_CAN_RXOVF+PortCANLast,
	Var_Last = Var_CAN_RXERR+PortCANLast,
}Var_t;
typedef char VarCheck[(NVM_Last==Var_Last ) ? 0 : -1];

static param_t parameters[Var_Last] = {{ (uint8_t*) "FORMATID", EEFORMAT, EEFORMAT, EEFORMAT, EEFORMAT},																	// EEFORMAT
		{ (uint8_t*) "NETID0", DEF_NODEID, 0, 255, DEF_NODEID},																										 														// NETID0
		{ (uint8_t*) "NETID1", DEF_NODEID+1, 0, 255, DEF_NODEID+1},																										 														// NETID1
		{ (uint8_t*) "NETID2", DEF_NODEID+2, 0, 255, DEF_NODEID+2},																										 														// NETID2
		{ (uint8_t*) "CAN2PWM_CH1", 1, 0, 16, 0 },																																																	// CH1
		{ (uint8_t*) "CAN2PWM_CH2", 2, 0, 16, 0 },																																															    // CH2
		{ (uint8_t*) "CAN2PWM_CH3", 3, 0, 16, 0 },                                                                                                  // CH3
		{ (uint8_t*) "CAN2PWM_CH4", 4, 0, 16, 0 },                                                                                                  // CH4
		{ (uint8_t*) "CAN2PWM_CH5", 5, 0, 16, 0 },                                                                                                  // CH5
		{ (uint8_t*) "CAN2PWM_CH6", 6, 0, 16, 0 },                                                                                                  // CH6
		{ (uint8_t*) "CAN2PWM_ACT_ESC1", 1, 0, 1, 0 },                                                                                              // CH1
		{ (uint8_t*) "CAN2PWM_ACT_ESC2", 1, 0, 1, 0 },                                                                                              // CH2
		{ (uint8_t*) "CAN2PWM_ACT_ESC3", 1, 0, 1, 0 },                                                                                              // CH3
		{ (uint8_t*) "CAN2PWM_ACT_ESC4", 1, 0, 1, 0 },                                                                                              // CH4
		{ (uint8_t*) "CAN2PWM_ACT_ESC5", 0, 0, 1, 0 },                                                                                              // CH5
		{ (uint8_t*) "CAN2PWM_ACT_ESC6", 0, 0, 1, 0 },
		{ (uint8_t*) "CAN1_TotBlocks ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN2_TotBlocks ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN3_TotBlocks ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN1_CurrBlocks", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN2_CurrBlocks", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN3_CurrBlocks", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN1_PeakBlocks", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN2_PeakBlocks", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN3_PeakBlocks", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN1_RXOVF	   ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN2_RXOVF	   ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN3_RXOVF	   ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN1_RXERR     ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN2_RXERR     ", 0, 0,65535, 0 },
		{ (uint8_t*) "CAN3_RXERR     ", 0, 0,65535, 0 },
};

static int8_t ActLookup[17] = { [0 ... sizeof(ActLookup) - 1]=-1 };							//  a lookup table to translate act id to pwm output
static int8_t EscLookup[20] = { [0 ... sizeof(EscLookup) - 1]=-1 };							//  a lookup table to translate esc index to pwm output
const rcpwm_t rcpwm_outputs[] = { { TIM5, 1, "TIM5_CH1_PA0" },                                                                                  // CH1
		{ TIM5, 2, "TIM5_CH2_PA1" },                                                                                                                // CH2
		{ TIM5, 3, "TIM5_CH3_PA2" },                                                                                                                // CH3
		{ TIM5, 4, "TIM5_CH4_PA3" },																																																								// CH4
		{ TIM2, 1, "TIM2_CH1_PA5" },                                                                                																// CH5
		{ TIM2, 3, "TIM2_CH3_PB10" },                                                                                                              	// CH6
};
uint64_t UniqID;
static bool IgnoreCanPort[PortCANLast]={0};
#if RFDCANLINK
#define MAX_NODES 20
#define MAX_TIDS	30
static int8_t PortLookup[PortCANLast][MAX_NODES]={[0 ... PortCANLast-1][0 ... MAX_NODES-1]=-1};
static tidEntry_t IDLookup[MAX_TIDS];
#endif

const Msg_Filt_t Msg_Filt[] = {
		{UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_ID,UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_SIGNATURE,CanardTransferTypeBroadcast},
		{UAVCAN_PROTOCOL_NODESTATUS_ID                ,UAVCAN_PROTOCOL_NODESTATUS_SIGNATURE                ,-1},
		{UAVCAN_PROTOCOL_GETNODEINFO_ID               ,UAVCAN_PROTOCOL_GETNODEINFO_SIGNATURE               ,-1},
		{RFD_AF3_IGNORECANPORT_ID                     ,RFD_AF3_IGNORECANPORT_SIGNATURE                     ,-1},
		{UAVCAN_PROTOCOL_PARAM_GETSET_ID              ,UAVCAN_PROTOCOL_PARAM_GETSET_SIGNATURE              ,-1},
#if RFDARBITER || RFDCANLINK
		{RFD_AF3_FLIGHTCONTROLLERHEALTH_ID            ,RFD_AF3_FLIGHTCONTROLLERHEALTH_SIGNATURE            ,-1},
		{RFD_AF3_VISIBLEFLIGHTCONTROLLER_ID           ,RFD_AF3_VISIBLEFLIGHTCONTROLLER_SIGNATURE           ,-1},
#endif
#if RFDENDPOINT || RFDCANLINK
		{RFD_AF3_USEFLIGHTCONTROLLER_ID               ,RFD_AF3_USEFLIGHTCONTROLLER_SIGNATURE               ,-1},
		{UAVCAN_EQUIPMENT_ESC_RAWCOMMAND_ID           ,UAVCAN_EQUIPMENT_ESC_RAWCOMMAND_SIGNATURE           ,-1},
#endif
#if RFDENDPOINT
		{UAVCAN_EQUIPMENT_ACTUATOR_ARRAYCOMMAND_ID    ,UAVCAN_EQUIPMENT_ACTUATOR_ARRAYCOMMAND_SIGNATURE    ,-1},
#endif
};

// ******************** local function prototypes ******************************
static void updateLookups(void);
static void sendCanard(PortCAN_t Port);
static void receiveCanard(PortCAN_t Port);
static void spinCanard(PortCAN_t Port);
static uint32_t getUptime(void);
static void getsetHandleCanard(CanardInstance* ins, CanardRxTransfer* transfer);

static bool shouldAcceptTransfer(const CanardInstance* ins, uint64_t* out_data_type_signature, uint16_t data_type_id,
		CanardTransferType transfer_type, uint8_t source_node_id);
#if RFDARBITER
static void UpdateHealth(PortCAN_t Port);
static void PubLishUseFC(PortCAN_t Port);
#elif RFDENDPOINT
static void UpdateVisibility(void);
static void PublishVisibility(PortCAN_t Port);
static void UpdateFCList(void);
static void rcpwmUpdate(float ar[6]);
static void DisablePWM(void);
static void rawcmdHandleCanard(CanardRxTransfer* transfer);
#endif
static void onTransferReceived(CanardInstance* ins, CanardRxTransfer* transfer, PortCAN_t Port);
static void onTransferReceived0(CanardInstance* ins, CanardRxTransfer* transfer);
static void onTransferReceived1(CanardInstance* ins, CanardRxTransfer* transfer);
static void onTransferReceived2(CanardInstance* ins, CanardRxTransfer* transfer);
static inline uint32_t getUptime(void);
void MainTask(void const * argument);
// ********************* Implementation ****************************************

bool shouldAcceptTransfer(const CanardInstance* ins, uint64_t* out_data_type_signature, uint16_t data_type_id, CanardTransferType transfer_type,
		uint8_t source_node_id)
{
  (void)source_node_id;

  if (canardGetLocalNodeID(ins) == CANARD_BROADCAST_NODE_ID)
  {
    //If we're in the process of allocation of dynamic node ID, accept only relevant transfers.
		if ((transfer_type == CanardTransferTypeBroadcast) &&	(data_type_id == UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_ID))
		{
			*out_data_type_signature = UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_SIGNATURE;
			return true;
		}
  }
  else
  {
  	for(uint16_t idx=0;idx<ARRAY_SIZE(Msg_Filt);idx++)
  	{
  		if(Msg_Filt[idx].transfer_type < 0 && Msg_Filt[idx].data_type_id == data_type_id)
  		{
  			*out_data_type_signature = Msg_Filt[idx].out_data_type_signature;
  			return(true);
  		}
  	}
  }
  return(false);
}
////////////////////////////////////////////////////////////////////////////////
static void onTransferReceived0(CanardInstance* ins, CanardRxTransfer* transfer)
{
	onTransferReceived(ins, transfer, 0);
}
static void onTransferReceived1(CanardInstance* ins, CanardRxTransfer* transfer)
{
	onTransferReceived(ins, transfer, 1);
}
static void onTransferReceived2(CanardInstance* ins, CanardRxTransfer* transfer)
{
	onTransferReceived(ins, transfer, 2);
}
////////////////////////////////////////////////////////////////////////////////
#if RFDCANLINK
static bool CANPortUsed(int8_t Port)
{
	bool Used=false;
	for(int8_t i=0;i<MAX_NODES&&!Used;i++)
	{
		if(PortLookup[Port][i] >=0){Used=true;}
	}
	return(Used);
}
static int8_t FindCANPort(uint8_t NodeID,int8_t *idx)
{
	int8_t Port=-1;
	for(int8_t p=0;p<PortCANLast&&Port<0;p++)
	{
		for(int8_t i=0;i<MAX_NODES&&Port<0;i++)
		{
			if(PortLookup[p][i] == NodeID){Port = p;if(idx)*idx=i;}
		}
	}
	return(Port);
}
////////////////////////////////////////////////////////////////////////////////
static bool AddCANPortLookup(uint8_t Port,uint8_t NodeID)
{
	bool Res =false;
	int8_t p,idx;
	if(Port>=PortCANLast){return(false);}
	if((p=FindCANPort(NodeID,&idx)) == Port)return(true);
	if(p >=0){PortLookup[p][idx]=-1;}					// clear old entry if it exists
	for(int8_t i=0;i<MAX_NODES&&!Res;i++)
	{
		if(PortLookup[Port][i] < 0){PortLookup[Port][i] = NodeID;Res=true;}
	}
	return(Res);
}
////////////////////////////////////////////////////////////////////////////////
static bool AddCANTIDLookup(uint8_t NodeID,uint8_t dtype_id,uint8_t *tid)				// add a new tid entry
{
	bool Res =false;
	for(int8_t i=0;i<ARRAY_SIZE(IDLookup)&&!Res;i++)															// go thru array
	{
		if(IDLookup[i].NodeID < 0)																									// if entry is not used
		{	IDLookup[i].dtype_id=dtype_id;IDLookup[i].NodeID=NodeID;									// assign entry details
		 	 tid=&IDLookup[i].transfer_id;IDLookup[i].TickmS=getUptime();Res=true;		// and record time
		}
	}
	return(Res);																																	// return true if added ok
}
////////////////////////////////////////////////////////////////////////////////
static uint8_t* FindCANTID(uint8_t NodeID,uint8_t dtype_id)											// add entry if it doesn't exist and return pointer to valid tid
{
	static uint32_t LastTickmS=0;
	uint8_t* TID=NULL;
	for(int8_t i=0;i<ARRAY_SIZE(IDLookup)&&!TID;i++)															// go thru array
	{
		if((IDLookup[i].NodeID == NodeID)&&(IDLookup[i].dtype_id == dtype_id))			// if nodeid and data type match
		{TID = &IDLookup[i].transfer_id;IDLookup[i].TickmS=getUptime();}						// set return pointer, and record activity time
	}
	if(!TID)																																			// if nothing found
	{
		AddCANTIDLookup(NodeID,dtype_id,TID);																				// try add a new entry
	}
	if ((getUptime() - LastTickmS) > TID_CHECK_MS)																// if time to check activity
	{
		LastTickmS = getUptime();																										// record new check time
		for(int8_t i=0;i<ARRAY_SIZE(IDLookup);i++)																	// go thru array
		{	if(IDLookup[i].NodeID >= 0)																								// if used entry
			{	if(LastTickmS-IDLookup[i].TickmS > TID_TIMEOUT_MS)											// if time for activity has expired
				{IDLookup[i].NodeID = -1;}																							// remove entry
			}
		}
	}
	return(TID);																																	// return the transfer id entry
}
#endif
////////////////////////////////////////////////////////////////////////////////
static void onTransferReceived(CanardInstance* ins, CanardRxTransfer* transfer, PortCAN_t Port)
{
#if RFDCANLINK
	bool Res;
	int len;
	if((transfer->transfer_type == CanardTransferTypeBroadcast)&&(transfer->data_type_id == UAVCAN_PROTOCOL_NODESTATUS_ID))
	{AddCANPortLookup(Port,transfer->source_node_id);}
	len = transfer->payload_len;
	Res=StartPacket(sizeof(CANRxTransfer)+len);
	if(Res)
	{
		CANRxTransfer tfer={
				.data_type_id = transfer->data_type_id,	.transfer_type = transfer->transfer_type,
				.transfer_id = transfer->transfer_id,		.priority = transfer->priority,
				.source_node_id = transfer->source_node_id,.dest_node_id = transfer->dest_node_id,
				.data_type_signature = transfer->data_type_signature};
		Res = AppendPacket((uint8_t*)&tfer,sizeof(tfer));
	}
	uint16_t offset=0;
	while(Res&&len)
	{
		static uint8_t *ptr;
		if((len=getDescatterTransferPayloadPtr(transfer,offset,&ptr)))
		{
			Res = AppendPacket(ptr,len);
			offset += len;
		}
	}
	if(Res){SendPacket();}
  if(transfer->dest_node_id && transfer->dest_node_id != canardGetLocalNodeID(ins))return;	// this is not the droid you are looking for
#endif
	// note messages can have the same ID but may be request response or broadcast
	if((transfer->transfer_type == CanardTransferTypeBroadcast)&&(transfer->data_type_id == UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_ID))
	{
		onDynNodeRx(ins,transfer,Port);
	}
	else if((transfer->transfer_type == CanardTransferTypeRequest)&&(transfer->data_type_id == UAVCAN_PROTOCOL_GETNODEINFO_ID))
	{
		onGetNodeInfoRx(ins,transfer,Port);
	}
	else if (transfer->data_type_id == UAVCAN_PROTOCOL_PARAM_GETSET_ID)
	{
		getsetHandleCanard(ins, transfer);
	}
	else if (transfer->data_type_id == RFD_AF3_IGNORECANPORT_ID)
	{
		rfd_af3_IgnoreCANPort dest;
		rfd_af3_IgnoreCANPort_decode_internal(transfer, 0, &dest, NULL, 0);
		if(dest.NodeID == gNodeID[Port])
		{
			IgnoreCanPort[0] =  dest.IgnoreCANPort[0];
			IgnoreCanPort[1] =  dest.IgnoreCANPort[1];
			IgnoreCanPort[2] =  dest.IgnoreCANPort[2];
		}
	}
#if RFDARBITER
	else if(transfer->data_type_id == RFD_AF3_VISIBLEFLIGHTCONTROLLER_ID)
	{
#if 0
		static volatile uint32_t tick,count =0,min=2000,max=0;
		if(++count >= 10)
		{
			count=0;
			volatile uint32_t ticks =getUptime() -tick;
			if(tick)
			{
				min = MIN(min,ticks);max = MAX(max,ticks);
				if(ticks < 950 ||  ticks > 1000)
				{__ASM volatile("BKPT #01");}
			}
			tick = ticks + tick;
		}
#endif
		if(IgnoreCanPort[Port])return;
		rfd_af3_VisibleFlightController dest;
		rfd_af3_VisibleFlightController_decode_internal(transfer, 0, &dest, NULL, 0);
		onVisibleFCRX(dest.CANPresent,dest.UniqID,getUptime());
	}
	else if (transfer->data_type_id == RFD_AF3_FLIGHTCONTROLLERHEALTH_ID)
	{
		if(IgnoreCanPort[Port])return;
		rfd_af3_FlightControllerHealth dest;
		rfd_af3_FlightControllerHealth_decode_internal(transfer, 0, &dest, NULL, 0);
		onPixHealtRX(Port,dest.ErrorScore,getUptime());
		//printf("Health[%i]: %li\n",Port,(int32_t)(1000*dest.ErrorScore));
	}
#elif RFDENDPOINT
	// these 2 commands are channel data look up parameters to see which goes where
	else if (transfer->data_type_id == UAVCAN_EQUIPMENT_ACTUATOR_ARRAYCOMMAND_ID)
	{
		if(IgnoreCanPort[Port])return;
		if(ActiveFCPort == Port)
		{rawcmdHandleCanard(transfer);}
		RxFCTick[Port] = getUptime();
		PixVisibility[Port] = true;
	}
	else if (transfer->data_type_id == UAVCAN_EQUIPMENT_ESC_RAWCOMMAND_ID)
	{
		if(IgnoreCanPort[Port])return;
		if(ActiveFCPort == Port)
		{rawcmdHandleCanard(transfer);}
		RxFCTick[Port] = getUptime();
		PixVisibility[Port] = true;
	}
	else if (transfer->data_type_id == RFD_AF3_USEFLIGHTCONTROLLER_ID)
	{
		if(IgnoreCanPort[Port])return;
		rfd_af3_UseFlightController dest;
		rfd_af3_UseFlightController_decode_internal(transfer, 0, &dest, NULL, 0);
		ActiveFCPort = UpdateVoter(Port, dest.ThisBus, dest.UseFConBus, dest.UniqID,getUptime());
		if(ActiveFCPort<0){DisablePWM();}
		//printf("ActivePort: %i",ActiveFCPort);
	}
#endif	
}
////////////////////////////////////////////////////////////////////////////////

int main_canard(void)
{
#if RFDARBITER
#ifdef DEBUG
	//ArbiterTest();
#endif
#elif RFDCANLINK
	PacketHandler_Init();
#endif
	InitNVM();
	for(int i=0;i<ARRAY_SIZE(parameters);i++)
	{
		NVM_GET(i,parameters[i].val);
	}
	PortCAN_t Port;
	gpioSet(LED1_GPIO_PORT, LED1_PIN,1);
	CanardSTM32AcceptanceFilterConfiguration filter_configs[ARRAY_SIZE(Msg_Filt)] =
		{[0 ... ARRAY_SIZE(Msg_Filt)-1].mask=CANARD_CAN_FRAME_RTR};

	for(uint16_t idx=0;idx<ARRAY_SIZE(filter_configs);idx++)
	{
		filter_configs[idx].id = Msg_Filt[idx].data_type_id;
	}
	InitSerial();
	gpioSet(CANEN_GPIO_PORT, CAN1EN_PIN,1);
	gpioSet(CANEN_GPIO_PORT, CAN2EN_PIN,1);
	gpioSet(CANEN_GPIO_PORT, CAN3EN_PIN,1);
	for (Port = CANStrt; Port < CANEnd; Port++)
	{
		CANInit((CAN_Init_t*) &CAN_Init[Port]);
	}
	for (Port = CANStrt; Port < CANEnd; Port++)
	{
		//gpioSet(LED2_GPIO_PORT, LED2_PIN,1);
		gNodeID[Port] = parameters[Port+Var_NETID0].val;
		CANswInit(Port, &g_canard[Port], g_canard_memory_pool[Port], sizeof(g_canard_memory_pool[0]),
				(Port == 0) ? (onTransferReceived0) : ((Port == 1) ? (onTransferReceived1) : (onTransferReceived2)), shouldAcceptTransfer, gNodeID[Port],
						filter_configs,ARRAY_SIZE(filter_configs));
		//gpioSet(LED3_GPIO_PORT, LED3_PIN,1);
	}
	gpioSet(CANEN_GPIO_PORT, CAN1EN_PIN,0);
	gpioSet(CANEN_GPIO_PORT, CAN2EN_PIN,0);
	gpioSet(CANEN_GPIO_PORT, CAN3EN_PIN,0);

	PWMInit();
	SysTick_Config(SystemCoreClock / 1000);
	osThreadDef(Task01, MainTask, osPriorityNormal, 0, 10*configMINIMAL_STACK_SIZE);
	MainTaskHandle = osThreadCreate(osThread(Task01), NULL);

	osKernelStart();
	return (0);
}
////////////////////////////////////////////////////////////////////////////////
#if RFDENDPOINT
static void updateLED(int8_t Port)
{
	int8_t val=0;
	if(Port>=0)
	{
		val=1<<Port;
	}
	gpioSet(LED1_GPIO_PORT, LED1_PIN,val&0x01);
	gpioSet(LED2_GPIO_PORT, LED2_PIN,val&0x02);
	gpioSet(LED3_GPIO_PORT, LED3_PIN,val&0x04);
}
#endif
////////////////////////////////////////////////////////////////////////////////
void MainTask(void const * argument)
{
	PortCAN_t Port;
	bool NodeValid[PortCANLast];
	gpioSet(LED3_GPIO_PORT, LED3_PIN,1);
	for(Port=0;Port<PortCANLast;Port++)
	{
		if(gNodeID[Port] != 0)
		{
			NodeValid[Port] = true;
			SetNodeMode(Port,UAVCAN_PROTOCOL_NODESTATUS_MODE_OPERATIONAL);
		}
	}
	Port=0;
	UniqID = LL_GetUID_Word1();
	UniqID = UniqID << 32;
	UniqID += LL_GetUID_Word0();
	UniqID ^= LL_GetUID_Word2();
	updateLookups();
	gpioSet(LED1_GPIO_PORT, LED1_PIN,0);
	gpioSet(LED2_GPIO_PORT, LED2_PIN,0);
	gpioSet(LED3_GPIO_PORT, LED3_PIN,0);
	while (1)
	{
#if RFDCANLINK
		int8_t port;
		static uint8_t SerBuff[1024];
		uint16_t Serlen = sizeof(SerBuff);
		if(Packet_Poll(SerBuff,&Serlen))
		{
			uint16_t idx=0;
			CANRxTransfer* tfer;
			tfer = (CANRxTransfer*)&SerBuff[idx];
			idx += sizeof(CANRxTransfer);
			int16_t len = Serlen-idx;
			uint8_t *tid = FindCANTID(tfer->source_node_id,tfer->data_type_id);
			if(tid&&(len >= 0))
			{
				*tid = tfer->transfer_id;
				if(tfer->transfer_type == CanardTransferTypeBroadcast)
				{
					for(int8_t p=0;p<PortCANLast;p++)
					{
						if(CANPortUsed(p))
						{
							canardBroadcast_internal(&g_canard[p],tfer->data_type_signature,tfer->data_type_id,
								tid,tfer->priority,&SerBuff[idx], len,tfer->source_node_id);
						}
					}
				}
				else if((tfer->transfer_type == CanardTransferTypeRequest)||(tfer->transfer_type == CanardTransferTypeResponse))
				{
					if((port=FindCANPort(tfer->dest_node_id,NULL)) >=0)
					{
						canardRequestOrRespond_internal(&g_canard[port],tfer->dest_node_id,tfer->data_type_signature,
								tfer->data_type_id,tid,tfer->priority,tfer->transfer_type,&SerBuff[idx],len,tfer->source_node_id);
					}
				}
			}
		}
#endif
		if(!NodeValid[Port])
		{
			if(DynAllocNodeID(&g_canard[Port],&gNodeID[Port],Port,sendCanard,receiveCanard))
			{
				SetNodeMode(Port,UAVCAN_PROTOCOL_NODESTATUS_MODE_OPERATIONAL);
				NodeValid[Port] = true;
			}
		}
		else
		{
			sendCanard(Port);
			receiveCanard(Port);
			spinCanard(Port);
#if RFDARBITER
			UpdateHealth(Port);
			if(!IgnoreCanPort[Port])
			{PubLishUseFC(Port);}
#elif RFDENDPOINT
			updateLED(ActiveFCPort);
			UpdateVisibility();																													// check the visibility matrix
			if(!IgnoreCanPort[Port])
			{PublishVisibility(Port);}																									 	// send the visibility message if required
			UpdateFCList();																															// check the arbiter active FCList and delete any expired entries
#endif
		//osDelay(100);
		}
		if (++Port >= CANEnd)
		{
			Port = CANStrt;
		}
		static uint32_t del=0;
		if(++del >= 10000UL)
		{
			static bool tog=0;
			gpioSet(LED3_GPIO_PORT, LED3_PIN,tog);
			tog =!tog;
			del=0;
		}

	}
}
////////////////////////////////////////////////////////////////////////////////
#if 0
void assert_failed(char* file, uint32_t line)
{
	__ASM volatile("BKPT #01");
	while(1);
}
#endif
////////////////////////////////////////////////////////////////////////////////
static void sendCanard(PortCAN_t Port)
{
	const CanardCANFrame* txf;
	static uint32_t txbusy[PortCANLast] = { 0 };
	static uint32_t TxStart[PortCANLast] = { 0 };
	static uint32_t CleanUp[PortCANLast] = { 0 };
	int tx_res;
	const uint32_t MAXTXMS = 1000;
	volatile uint32_t tnow = getUptime();
	if((tnow - CleanUp[Port])> (CANARD_RECOMMENDED_STALE_TRANSFER_CLEANUP_INTERVAL_USEC/1000))
	{
		CleanUp[Port] = tnow;
		canardCleanupStaleTransfers(&g_canard[Port], getUptime() * 1000);
	}

	if (txbusy[Port])
	{
		if ((tnow - TxStart[Port]) > MAXTXMS)
		{
			//printf("Port[%d] tx failed\n",Port+1);
			while ((txf = canardPeekTxQueue(&g_canard[Port])))
			{
				canardPopTxQueue(&g_canard[Port]);
			}
			canardCleanupStaleTransfers(&g_canard[Port], getUptime() * 1000);
			txbusy[Port] = 0;
		}
	}
	txf = canardPeekTxQueue(&g_canard[Port]);
	if (txf)
	{
#if RFDARBITER||RFDCANLINK
		gpioToggle(LED1_GPIO_PORT, LED1_PIN);
#endif
		tx_res = canardSTM32Transmit(txf, Port);
		if (tx_res < 0)                  																						// Failure - drop the frame and report
		{
			__ASM volatile("BKPT #01");																								// TODO: handle the error properly
		} else if (tx_res > 0)
		{
			txbusy[Port] = 0;
			canardPopTxQueue(&g_canard[Port]);
		} else if (!txbusy[Port])
		{
			TxStart[Port] = getUptime();
			txbusy[Port] = 1;
		}
	}
}
////////////////////////////////////////////////////////////////////////////////
static void receiveCanard(PortCAN_t Port)
{
	CanardCANFrame rx_frame;
	int res = canardSTM32Receive(&rx_frame, Port);
	if (res)
	{
#if RFDARBITER||RFDCANLINK
		gpioToggle(LED2_GPIO_PORT, LED2_PIN);
#endif
#if RFDCANLINK
		canardHandleRxFrame_internal(&g_canard[Port], &rx_frame, getUptime() * 1000,true);
#else
		canardHandleRxFrame(&g_canard[Port], &rx_frame, getUptime() * 1000);
#endif
	}
}
////////////////////////////////////////////////////////////////////////////////
static void spinCanard(PortCAN_t Port)
{
	static uint32_t spin_time[PortCANLast] = { 0 };
	if (getUptime() < spin_time[Port] + CANARD_SPIN_PERIOD) return; 							// rate limiting
	spin_time[Port] = getUptime();
	uint8_t buffer[UAVCAN_PROTOCOL_NODESTATUS_MAX_SIZE];
	static uint8_t transfer_id[PortCANLast] = {0}; 																							// This variable MUST BE STATIC; refer to the libcanard documentation
	uavcan_protocol_NodeStatus source = { 0 };
	source.health = UAVCAN_PROTOCOL_NODESTATUS_HEALTH_OK;
	source.mode = UAVCAN_PROTOCOL_NODESTATUS_MODE_OPERATIONAL;
	source.uptime_sec = (getUptime() / 1000);
	uint16_t len = uavcan_protocol_NodeStatus_encode(&source, buffer);
	canardBroadcast(&g_canard[Port],
	UAVCAN_PROTOCOL_NODESTATUS_SIGNATURE,
	UAVCAN_PROTOCOL_NODESTATUS_ID, &transfer_id[Port],
	CANARD_TRANSFER_PRIORITY_LOW, buffer, len);
}
////////////////////////////////////////////////////////////////////////////////
static inline uint32_t getUptime(void)
{
	return (osKernelSysTick());
}
////////////////////////////////////////////////////////////////////////////////
static void updateLookups(void)
{
	int8_t i;
	for (i = 0; i < ARRAY_SIZE(ActLookup); i++)
	{
		ActLookup[i] = -1;
	}
	for (i = 0; i < ARRAY_SIZE(EscLookup); i++)
	{
		EscLookup[i] = -1;
	}
	for (i = 0; i < NUM_PWM; i++)
	{
		PWMSetFreq(rcpwm_outputs[i].timer,rcpwm_outputs[i].channel,PWM_Freq50Hz);//
		if (parameters[i+Var_CAN2PWM_CHStart].val)																	// if a valid output
		{
			if (0 == parameters[i + Var_CAN2PWM_ACTStart].val)												// it's an actuator
			{
				ActLookup[parameters[i + Var_CAN2PWM_CHStart].val] = i;									// set lookup to this index
			} else 																																		// it's an esc
			{
				EscLookup[parameters[i+Var_CAN2PWM_CHStart].val - 1] = i;								// set lookup to this index
				PWMSetFreq(rcpwm_outputs[i].timer,rcpwm_outputs[i].channel,PWM_Freq490Hz);
			}
		}
	}
}
////////////////////////////////////////////////////////////////////////////////
#if RFDARBITER
static void UpdateHealth(PortCAN_t Port)
{
	static uint32_t time[PortCANLast]= {0};
	uint32_t now = getUptime();
	if (now < time[Port] + HEALTH_UPDATE_MS)																			// rate limiting
	{
		return;
	}
	time[Port] = now;
	UpdateHealth50mS(Port,now);
}
////////////////////////////////////////////////////////////////////////////////
static void PubLishUseFC(PortCAN_t Port)
{
	static uint32_t publish_time[PortCANLast] = { 0 };
	if (getUptime() < publish_time[Port] + USEFC_PUB_MS)
	{
		return;
	} 																																						// rate limiting
	publish_time[Port] = getUptime();
	uint8_t buffer[RFD_AF3_USEFLIGHTCONTROLLER_MAX_SIZE];
	static uint8_t transfer_id[PortCANLast] = {0};
	rfd_af3_UseFlightController source;
	source.ThisBus = Port;
	source.UseFConBus = FindBestFCPort();
	source.UniqID = UniqID;
	memset(buffer, 0x00, RFD_AF3_USEFLIGHTCONTROLLER_MAX_SIZE);
	uint32_t length = rfd_af3_UseFlightController_encode(&source, buffer);
	canardBroadcast(&g_canard[Port],
	RFD_AF3_USEFLIGHTCONTROLLER_SIGNATURE,
	RFD_AF3_USEFLIGHTCONTROLLER_ID, &transfer_id[Port],
	CANARD_TRANSFER_PRIORITY_LOW, &buffer[0], length);
}
#elif RFDENDPOINT
static void rawcmdHandleCanard(CanardRxTransfer* transfer)
{
	static float Level[6];																												// -1 to +1
																																								// can be an esc command or actuator
																																								// must use parameters to find out which PWM out to put channel data
	if (UAVCAN_EQUIPMENT_ACTUATOR_ARRAYCOMMAND_ID == transfer->data_type_id)
	{
		int32_t act[15] = { 0 };
		uint8_t* DynArr = (uint8_t*) act;
		uavcan_equipment_actuator_ArrayCommand dest;
		uavcan_equipment_actuator_ArrayCommand_decode_internal(transfer, transfer->payload_len, &dest, &DynArr, 0);
		uavcan_equipment_actuator_Command * ptr = dest.commands.data;
		for (int i = 0; i < dest.commands.len; i++)
		{
			if (UAVCAN_EQUIPMENT_ACTUATOR_COMMAND_COMMAND_TYPE_UNITLESS == ptr[i].command_type)
			{
				int8_t index = ActLookup[ptr[i].actuator_id];
				if (index >= 0)
				{
					Level[index] = ptr[i].command_value;
				}
			}
		}
	} else if (UAVCAN_EQUIPMENT_ESC_RAWCOMMAND_ID == transfer->data_type_id)
	{
		int16_t esc[20] = { 0 };
		uint8_t* DynArr = (uint8_t*) esc;
		uavcan_equipment_esc_RawCommand dest;
		uavcan_equipment_esc_RawCommand_decode_internal(transfer, transfer->payload_len, &dest, &DynArr, 0);
		if (dest.cmd.len > ARRAY_SIZE(EscLookup))
		{
			dest.cmd.len = ARRAY_SIZE(EscLookup);
		}
		for (int i = 0; i < dest.cmd.len; i++)
		{
			if (EscLookup[i] >= 0)
			{
				int16_t Val = LIMIT(dest.cmd.data[i],ESC_RAWCOMMAND_MIN_VALUE,ESC_RAWCOMMAND_MAX_VALUE);
				Val -= ESC_RAWCOMMAND_MIN_VALUE;
				Level[EscLookup[i]] = ((float) Val) / ESC_RAWCOMMAND_RANGE;
			}
		}
	}
	rcpwmUpdate(Level);
}
////////////////////////////////////////////////////////////////////////////////
static void DisablePWM(void)
{
	for (uint8_t i = 0; i < 6; i++)
	{
		TIM_SetCompare(rcpwm_outputs[i].timer, rcpwm_outputs[i].channel, 0);
	}
}
static void rcpwmUpdate(float ar[6])
{
	static float Lastar[6];
	bool diff = 0;//(0 != memcmp(Lastar, ar, sizeof(Lastar)));
	memcpy(Lastar, ar, sizeof(Lastar));
	if (diff)
	{
		printf("PWM: ");
	}
	for (uint8_t i = 0; i < 6; i++)
	{
		uint32_t val = RCPWM_MIN_uS;
		ar[i] = ar[i] * (float) RCPWM_RANGE_uS;
		val += ((int32_t) ar[i]);
		TIM_SetCompare(rcpwm_outputs[i].timer, rcpwm_outputs[i].channel, uS_to_ticks(val));
		if (diff)
		{
			printf("%lu ", val);
		}
	}
	if (diff)
	{
		printf("\n");
	}
}
////////////////////////////////////////////////////////////////////////////////
static void UpdateFCList(void)
{
	static uint32_t Last = 0;
	uint32_t Now =getUptime();
	if (Now < Last + CHECKFC_PERIOD_mS)
	{
		return;
	} 					// rate limiting
	Last = Now;
	UpdateVoterList50mS(&ActiveFCPort,Now);																			// returns true if active port has been updated
	if(ActiveFCPort<0){DisablePWM();}
}
////////////////////////////////////////////////////////////////////////////////
static void UpdateVisibility(void) 																							// check visibility matrix
{
	static uint32_t last= 0;
	uint32_t now = getUptime();
	if (now < last + VIS_CHECK_MS)
	{
		return;
	} 																																						// rate limiting
	last = now;
	for(PortCAN_t Port=0;Port<PortCANLast;Port++)
	{
		if(PixVisibility[Port])
		{
			if ((now - RxFCTick[Port]) >= VS_TIMEOUT_MS)															// if it has timed out
			{
				PixVisibility[Port] = 0;																								// invalidate health data
			}
		}
	}
}

static void PublishVisibility(PortCAN_t Port) 																	// send new visibility message to arbiter
{
	static uint32_t publish_time[PortCANLast] = { 0 };
	uint32_t now = getUptime();
	if (now < publish_time[Port] + VIS_PUB_MS)
	{
		return;
	} 																																						// rate limiting
	publish_time[Port] = now;
	uint8_t buffer[RFD_AF3_VISIBLEFLIGHTCONTROLLER_MAX_SIZE];
	static uint8_t transfer_id[PortCANLast] = {0};
	rfd_af3_VisibleFlightController source;
	source.UniqID = UniqID;
	source.CANPresent[0] = PixVisibility[0];
	source.CANPresent[1] = PixVisibility[1];
	source.CANPresent[2] = PixVisibility[2];
	memset(buffer, 0x00, RFD_AF3_VISIBLEFLIGHTCONTROLLER_MAX_SIZE);
	uint32_t length = rfd_af3_VisibleFlightController_encode(&source, buffer);
	canardBroadcast(&g_canard[Port],
	RFD_AF3_VISIBLEFLIGHTCONTROLLER_SIGNATURE,
	RFD_AF3_VISIBLEFLIGHTCONTROLLER_ID, &transfer_id[Port],
	CANARD_TRANSFER_PRIORITY_LOW, &buffer[0], length);
}
#endif
////////////////////////////////////////////////////////////////////////////////
static inline param_t * getParamByIndex(uint16_t index)
{
	if (index >= ARRAY_SIZE(parameters))
	{
		return NULL;
	}

	return &parameters[index];
}
////////////////////////////////////////////////////////////////////////////////
static inline param_t * getParamByName(uint8_t * name,uint8_t * idx)
{
	for (uint16_t i = 0; i < ARRAY_SIZE(parameters); i++)
	{
		if (strncmp((char const*) name, (char const*) parameters[i].name, strlen((char const*) parameters[i].name)) == 0)
		{
			*idx = i;
			return &parameters[i];
		}
	}
	return NULL;
}
////////////////////////////////////////////////////////////////////////////////
static void getsetHandleCanard(CanardInstance* ins, CanardRxTransfer* transfer)
{
	param_t * p = NULL;
	uint8_t idx=0;
	uint8_t name[UAVCAN_PROTOCOL_PARAM_GETSET_REQUEST_NAME_MAX_LENGTH] = "";
	uint8_t* DynArr = name;
	uavcan_protocol_param_GetSetRequest dest;
	uavcan_protocol_param_GetSetRequest_decode(transfer, transfer->payload_len, &dest, &DynArr);
	if (strlen((char const*) dest.name.data))
	{
		p = getParamByName(dest.name.data,&idx);
	} else
	{
		idx = dest.index;
		p = getParamByIndex(dest.index);
	}
	uint8_t buffer[UAVCAN_PROTOCOL_PARAM_GETSET_RESPONSE_MAX_SIZE];
	uavcan_protocol_param_GetSetResponse source = { 0 };
	if (p)
	{
		switch(idx)
		{
			case Var_CAN_TotBlocks +0:p->val = g_canard[0].allocator.statistics.capacity_blocks;break;
			case Var_CAN_TotBlocks +1:p->val = g_canard[1].allocator.statistics.capacity_blocks;break;
			case Var_CAN_TotBlocks +2:p->val = g_canard[2].allocator.statistics.capacity_blocks;break;
			case Var_CAN_CurrBlocks+0:p->val = g_canard[0].allocator.statistics.current_usage_blocks;break;
			case Var_CAN_CurrBlocks+1:p->val = g_canard[1].allocator.statistics.current_usage_blocks;break;
			case Var_CAN_CurrBlocks+2:p->val = g_canard[2].allocator.statistics.current_usage_blocks;break;
			case Var_CAN_PeakBlocks+0:p->val = g_canard[0].allocator.statistics.peak_usage_blocks;break;
			case Var_CAN_PeakBlocks+1:p->val = g_canard[1].allocator.statistics.peak_usage_blocks;break;
			case Var_CAN_PeakBlocks+2:p->val = g_canard[2].allocator.statistics.peak_usage_blocks;break;
			case Var_CAN_RXOVF	   +0:p->val = (canardSTM32GetStats(0)).rx_overflow_count;break;
			case Var_CAN_RXOVF	   +1:p->val = (canardSTM32GetStats(1)).rx_overflow_count;break;
			case Var_CAN_RXOVF	   +2:p->val = (canardSTM32GetStats(2)).rx_overflow_count;break;
			case Var_CAN_RXERR     +0:p->val = (canardSTM32GetStats(0)).error_count      ;break;
			case Var_CAN_RXERR     +1:p->val = (canardSTM32GetStats(1)).error_count      ;break;
			case Var_CAN_RXERR     +2:p->val = (canardSTM32GetStats(2)).error_count      ;break;
			default:break;
		}
		if(idx&&(dest.value.union_tag == UAVCAN_PROTOCOL_PARAM_VALUE_INTEGER_VALUE))// can't write the first var
		{
			p->val = dest.value.integer_value;
			updateLookups();																													// data has changed update the lookup tables
			NVMSet(idx,p->val);																												// write eevar
		}
		source.value.union_tag = UAVCAN_PROTOCOL_PARAM_VALUE_INTEGER_VALUE;
		source.max_value.union_tag = UAVCAN_PROTOCOL_PARAM_VALUE_INTEGER_VALUE;
		source.min_value.union_tag = UAVCAN_PROTOCOL_PARAM_VALUE_INTEGER_VALUE;
		source.default_value.union_tag = UAVCAN_PROTOCOL_PARAM_VALUE_INTEGER_VALUE;
		source.value.integer_value = p->val;
		source.min_value.integer_value = p->min;
		source.max_value.integer_value = p->max;
		source.default_value.integer_value = p->defval;
		source.name.data = p->name;
		source.name.len = strlen((char*) p->name);
	}
	uint16_t len = uavcan_protocol_param_GetSetResponse_encode(&source, buffer);
	int result = canardRequestOrRespond(ins, transfer->source_node_id,
	UAVCAN_PROTOCOL_PARAM_GETSET_SIGNATURE,
	UAVCAN_PROTOCOL_PARAM_GETSET_ID, &transfer->transfer_id, transfer->priority, CanardResponse, &buffer[0], (uint16_t) len);
	(void) result;
}
// ********************* main_canard.c *****************************************
